
#pragma once


///////////////////////////HASH CODE TABLE ////////////////////////////////////////////////

template <class T>
class HASHNAMED
{
public:
  char name[64]; // a method name should never be more than 64 bytes long
  T     data;

};

template <class T>
class HASHTABLEDEF {
public:
  HASHNAMED<T> **v;
  size_t size;
  size_t used;
  size_t usedLim;
};

template <class T>
class HASHTABLEITER {
  HASHNAMED<T> **p;
  HASHNAMED<T> **end;
};


template <class T>
class _HashTable
{

	// Members
protected:
	HASHTABLEDEF<T> m_hTable;
	bool			m_bCaseSensitive; // used when we try to match names

	// Constructor
public:
	_HashTable();

	// Members
public:
	void		Init(bool bCaseSensitive=false);
	void		Destroy();

	T			Lookup(const char *name, T data, bool bAdd=false);
	T			Add(const char *name, T data); // Lookup(..., bAdd=true)

	long		Size();
	HASHNAMED<T>*	FirstItem(HASHTABLEITER<T> *iter);
	HASHNAMED<T>*	NextItem(HASHTABLEITER<T> *iter);

	// Accessors
	void		SetCaseSensitive(bool bIsCaseSensitive);
	bool		GetCaseSensitive();

// internal
protected:
	int				keyeq(const char *s1, const char *s2);
	unsigned long	hash(const char *s);
};




typedef _HashTable<void*>	HashTableLPVOID;
typedef _HashTable<int>		HashTableINT;


////////////////////////// How to use the hash table ////////////////////////:
/////////////////////////////////////////////////////////////////////////////
//
// the hash table provides a mapping between a name and an item
// the name is necessarily a unique char* base type
// the item can be of any type, the reason wht the hash table has T as template type
//
// the hash table provides both semi-direct access to data
// and this one allows also a for(i=1 to n) access loop.
//
//  most of the time T is an int or a ptr to something
//
//  when T is an int, use HashTableINT.
//  when T is a ptr to something, use HashTableLPVOID.
//  otherwise use your own template.
//
//  Follows how to instantiate an hash table :
//
//	HashTableINT h;
//	h.Add("html",1);
//	h.Add("head",2);
//	h.Add("table",3);
//
//  And now follows how to check whether a given item is in the hash table :
//
//  int token = h.Lookup(szTag,NULL);
//
//  token return is NULL if not found
//
//  features of hash table :
//   - names are case sensitive or not
//  featurs to come :
//   - return values can be customized
//   - serialization




#define XMLINIT_SIZE 64



template <class T> _HashTable<T>::_HashTable()
{
	Init();
}

template <class T> int _HashTable<T>::keyeq(const char *s1, const char *s2)
{
  for (; *s1 == *s2; s1++, s2++)
    if (*s1 == 0)
      return 1;
  return 0;
}

template <class T> unsigned long _HashTable<T>::hash(const char *s)
{
  unsigned long h = 0;
  while (*s)
    h = (h << 5) + h + (unsigned char)*s++;
  return h;
}

template <class T> T _HashTable<T>::Lookup(const char *name, T data, bool bAdd)
{
  size_t i;
  if (m_hTable.size == 0) // non initialized table
  {
    m_hTable.v = (HASHNAMED<T>**) calloc(XMLINIT_SIZE, sizeof(HASHNAMED<T> *));
    if (!m_hTable.v)
      return NULL;
    m_hTable.size = XMLINIT_SIZE;
    m_hTable.usedLim = XMLINIT_SIZE / 2;
    i = hash(name) & (m_hTable.size - 1);

	if (!bAdd)
		return NULL;
  }
  else // find name in the table
  {
    unsigned long h = hash(name);
	HASHNAMED<T> *p = NULL;
    for (i = h & (m_hTable.size - 1); m_hTable.v[i]; i == 0 ? i = m_hTable.size - 1 : --i) 
	{
      if (keyeq(name, m_hTable.v[i]->name))
	  {
		p = m_hTable.v[i];
		break;
	  }
    }
	if (p)
	{
		if (bAdd) // set new value for ptr
			p->data = data;
		return p->data;
	}

	if (!bAdd)
		return NULL;

	// not found, prepare addition to the table
    if (m_hTable.used == m_hTable.usedLim) 
	{
      // check for overflow
      size_t newSize = m_hTable.size * 2;
      HASHNAMED<T> **newV = (HASHNAMED<T>**) calloc(newSize, sizeof(HASHNAMED<T> *));
      if (!newV)
		return NULL; // out of memory
      for (i = 0; i < m_hTable.size; i++)
		if (m_hTable.v[i]) 
		{
			size_t j;
			for (j = hash(m_hTable.v[i]->name) & (newSize - 1);
					newV[j];
					j == 0 ? j = newSize - 1 : --j)
				;

			newV[j] = m_hTable.v[i];
		}
      free(m_hTable.v);
      m_hTable.v = newV;
      m_hTable.size = newSize;
      m_hTable.usedLim = newSize/2;
      for (i = h & (m_hTable.size - 1);  m_hTable.v[i];  i == 0 ? i = m_hTable.size - 1 : --i)
			;
    }
  }

  m_hTable.v[i] = (HASHNAMED<T>*) calloc(1, sizeof(HASHNAMED<T>) );
  if (!m_hTable.v[i])
     return NULL;
  strcpy(m_hTable.v[i]->name, name);
  m_hTable.v[i]->data = data;
  (m_hTable.used)++;

  return data; // const to non const
}

template <class T> T _HashTable<T>::Add(const char *name, T data)
{
	return Lookup(name,data,true);
}


template <class T> void _HashTable<T>::Destroy()
{
  HASHNAMED *p;
  register long nSize = m_hTable.size;
  for (register long i = 0; i < nSize; i++) 
  {
    p = m_hTable.v[i];
    if (p)
      free(p);
  }

  if (m_hTable.v)
	free(m_hTable.v);

  // init again
  Init();

}

template <class T> void _HashTable<T>::Init(bool bCaseSensitive)
{
  m_hTable.size = 0;
  m_hTable.usedLim = 0;
  m_hTable.used = 0;
  m_hTable.v = NULL;
}

template <class T> long _HashTable<T>::Size()
{
	return m_hTable.used;
}


template <class T> HASHNAMED<T>* _HashTable<T>::FirstItem(HASHTABLEITER<T> *iter)
{
	if (!iter)
		return NULL;

	iter->p = m_hTable.v;
	iter->end = iter->p + m_hTable.size;

	return NextItem(iter);
}

template <class T> HASHNAMED<T>* _HashTable<T>::NextItem(HASHTABLEITER<T> *iter)
{
	while (iter->p != iter->end) 
	{
		HASHNAMED<T> *tem = *(iter->p)++;
		if (tem)
			return tem;
	}
	return NULL;
}

